/*	
 *		ELF Code Reader -- Dump2Code v0.2
 *
 *				coded by newbug[at]chroot.org
 */

#include <elf.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <string.h>

#define	DEFAULT_SECTION		".text"
#define	LINE			14

void err_exit(const char *msg)
{
	perror(msg);
	exit(1);
}

Elf32_Shdr
*save_section_header(unsigned char *start, Elf32_Half num, Elf32_Half size)
{       
        unsigned char *ret;

        ret = (unsigned char *)malloc(num*size);
        memcpy((void *)ret, (void *)start, num*size);
        return (Elf32_Shdr *)ret;
}

int section_size,section_addr;
int
search_section(Elf32_Shdr *shdr, char *section, int num, unsigned char *strtab)
{                                                                         
        Elf32_Shdr *p;
        int i;       
			             
        p = shdr;

	if(section == NULL)
		section = DEFAULT_SECTION;

        for(i=0;i<num;i++)
		if(!strcmp(strtab+(p+i)->sh_name, section))
		{
                        printf("[%d]%s\n", i, strtab+(p+i)->sh_name);
                        section_size = (p+i)->sh_size;
                        section_addr = (p+i)->sh_offset;
			return 1;
                }
	return 0;
}

int is_elf_file(unsigned char *p)
{       
        if(memcmp((void *)p, ELFMAG, strlen(ELFMAG)))
                return 0;
        return 1;
}       

void usage(char *arg)
{
	printf("dump2code v0.2  coded by newbug[at]chroot.org\n");
	printf("Usage: %s -f filename [-O output filename] [-s section name] \n \
			[-l length] [-o offset] [-0] [-h]\n \
			default_section = \".text\"\n", arg);
	exit(0);
}
void main(int argc,char **argv)
{
	FILE *fp=NULL;
	int fd;
	unsigned char *start;
	struct stat buf;
	Elf32_Ehdr ehdr;
	Elf32_Shdr *shdr;
	unsigned char *strtab;
	int i,len=-1,offset=0;
	char ch;
	char *file=NULL, *section=NULL;
	int zero = 0;   // zero flag
	
	if(argc == 1)
		usage(argv[0]);
	
	while((ch = getopt(argc, argv, "f:l:s:o:O:0h"))!=-1)
	{
		switch(ch)
		{
			case 'f':
				file = (char *)malloc(strlen(optarg));
				strcpy(file, optarg);
				break;
			case 'O':
				fp = fopen(optarg, "w");
				break;
			case 's':
				section = (char *)malloc(strlen(optarg));
				strcpy(section, optarg);
				break;
			case 'l':
				len = atoi(optarg);
				break;
			case 'o':
				offset = atoi(optarg);
				break;
			case '0':
				zero = 1;
				break;	
			case 'h':
			default:
				usage(argv[0]);
		}
	}
	fd = open(file, O_RDONLY);
	if( fd == -1)
		err_exit("open");
	fstat(fd, &buf);
	start = (unsigned char *)\
		mmap(0, buf.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
	close(fd);
	memcpy((void *)&ehdr, start, sizeof(ehdr));
	if(is_elf_file((unsigned char *)&ehdr))
	{
		/* Save Section header */
		shdr = (Elf32_Shdr *)save_section_header((unsigned char *) \
					start+ehdr.e_shoff, ehdr.e_shnum, \
			    		ehdr.e_shentsize);
		strtab = start+(shdr + ehdr.e_shstrndx)->sh_offset;

		if(search_section(shdr, section, ehdr.e_shnum, strtab) ==0)
			printf("Section %s not found.\n",section);
		else
		{
			if(fp == NULL)
				fp = stdout;
			if(len == -1)
				len = section_size;
			fprintf(fp, "char code[] = \n\t\"");
			for(i=0+offset;i<len+offset;i++)
			{
				if(zero && (*(start+section_addr+i)== 0x0))
					break;
				fprintf(fp, "\\x%.2x",*(start+section_addr+i));
				if(!((i+1)%LINE))
					fprintf(fp, "\"\n\t\"");
			}
			if((i+1)%LINE)
				fprintf(fp, "\"\n\t");
			fprintf(fp,";\n");
			fclose(fp);
		}
		free(shdr);
	}
	else
		printf("Not ELF ...\n");
	munmap(start, buf.st_size);

}
